/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors: Maurizio Munafo', Marco Mellia, Richard Lobb, Gloria Bogliolo, 
/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Maurizio Munafo', Marco Mellia, Gloria Bogliolo, 
Ana Paula Couto, Richard Lobb
Telecomunication Networks Group
Politecnico di Torino
Torino, Italy
mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Declares the AOPeer subclass of Peer.
	This peer continually adjusts its neighbourhood so as to maintain
	a specified fraction of its outlinks unused over a specified time
	window.
*/

#ifndef _AO_PEER_H_
#define _AO_PEER_H_

#include <vector>
#include "types.h"
#include "peer.h"
#include "propertymap.h"
#include "globals.h"

/** Enumeration of the two different algorithms used for neighbourhood replenishment */
enum ReplenishModeType {RANDOM_REPLENISH, BA_REPLENISH, BAD_VALUE};


class AOPeer : public Peer
{
public:
	enum SchedType {OUTDEG=0, SQRT_OUTDEG, UPLOAD_BW, SQRT_UPLOAD_BW};
	static SchedType cs_desirability;	///< Desirability function for chunk sched selection
private:
	static size_t scan_period;			///< Interval (in input chunks) between neighbourhood scans
	static size_t low_tide_percent;		///< Minimum percent of links unused per scan cycle
	static size_t high_tide_percent;	///< Maximum percent of links unused per scan cycle
	static ReplenishModeType replenish_mode;  ///< Mode for neighbourhood replenishment algorithm
	static size_t min_in_links;			///< The minimum number of in-links to be maintained
	static size_t min_out_links;		///< The minimum number of out-links to be maintained
	static double min_connect_time;		///< The minimum time (in chunks) for which a churning peer connects
	static double max_connect_time;		///< The maximum time (in chunks) for which a churning peer connects
	static SchedType n_desirability;	///< Desirability function for neighbour selection

	size_t num_scan_cycles;				///< Number of neighbour scans performed
	size_t chunks_since_last_scan;		///< Number of chunks received since last scan
	size_t chunks_out_since_last_scan;  ///< Number of chunks sent by this peer since last scan

public:
	AOPeer(
		Bandwidth upload_bandwidth,		///< The upload_bandwidth of the peer
		Bandwidth download_bandwidth,	///< The download_bandwidth of the peer
		size_t peer_class				///< The bandwidth class of this peer
	) : Peer(upload_bandwidth, download_bandwidth, peer_class), num_scan_cycles(0),
			chunks_out_since_last_scan(0)
	{
		chunks_since_last_scan = (size_t) uniform(0, scan_period, &gSeed); // Random phasing of scans
	};
	static void init_statics();			///< Initialise all class static variables
	virtual void schedule_new_chunk();	///< Schedule the sending of a new chunk if possible
	virtual void receive_chunk(PChunk chunk, PPeer from); ///< Receive a chunk from a nieghbour
	double desirability();			///< Estimate of how desirable this peer is as a neighbour
	void adjust_neighbourhood();	///< Adapt overlay
	///  Shrink the neighbourhood to target mid-tide by culling subset of unused_links
	void shrink_neighbourhood(size_t num_used, std::vector<PLink>& unused_links);
	void grow_neighbourhood(size_t num_wanted);  ///< Grow the neighbourhood by specified amount
	/// Select a desirable subset of neighbours from a given candidate list
	void stochastic_select(size_t num_wanted, PeerList& candidates, PeerList& selection);
	/// Discard all existing in and out links and reconnect to the overlay as per GNR
	void reset_all_links();
	void set_churn(bool churner); ///< Turn churn-mode on or off
};
#endif
