/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Maurizio Munafo', Marco Mellia, Gloria Bogliolo, Ana Paula Couto,
		  Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Implementations of the dynamically-reconfigurable "ConfigPeer"
	subclass of the Peer class.
 */

#include <sstream>
#include <cmath>
#include "configpeer.h"
#include "aopeer.h"
#include "misc.h"
#include "chunk.h"
#include "link.h"
#include "globals.h"
#include "random.h"
#include "overlay.h"

//=======================ConfigPeer ==============================
// This is a peer that can be dynamically re-configured over a range of
// chunk scheduling strategies. The strategy to be followed by the peer
// is set by a call to its set_strategy method. The parameter is a string
// which is a sequence of tokens setting various attributes of the chunk
// scheduling strategy. To understand this, first read the documentation in
// algorithms.h of the 'best' and 'weighted' search strategies, which defines
// how an entity of some type is chosen from a set of such entities using
// a filter function to select a subset then a weight function to weight the
// different elements. Roughly speaking, a "best" strategy chooses the
// candidate with the highest weight, while a "weighted" strategy chooses
// one probabilisitically according to the weight. The chunk scheduling
// strategies all use those algorithms to select chunks and peers. The
// different algorithms can be classified as "chunk first" if a chunk is
// selected first and then an appropriate neighbouring peer is chosen, or
// "peer first" if the peer is selected and then an appropriate chunk is
// chosen. The various sub-strategies use different weights and different
// filters, as specified by the configuration string, to filter and weight
// both chunks and peers.

// First define various peer and chunk selector and weight functions

/// True iff chunk is useful to neighbour on given link
bool useful_to_selected(PChunk chunk, const PLink link) {
	return link->to()->needs_chunk(chunk);
}

/// True iff chunk is useful to any of the neighbours given by the iterator pair
bool useful_to_any(PChunk chunk, std::pair<LinkIterator, LinkIterator> it_pair) {
	for (LinkIterator it = it_pair.first; it != it_pair.second; ++it) {
		PPeer to = (*it)->to();
		if (to->needs_chunk(chunk)) {
			return true;
		}
	}
	return false;
}

/// True iff neighbour is not currently busy downloading
bool not_busy(const PLink link) {
	return !link->to()->download_busy();
}

/// True iff neighbour on given link needs any of the chunks in the given window
/// and is not current busy.
bool needs_any_and_not_busy(const PLink link, const PWindow pwin) {
	PPeer neighbour = link->to();
	if (!neighbour->download_busy()) {
		for (ChunkIterator cit = pwin->begin(); cit != pwin->end(); ++cit) {
			if (neighbour->needs_chunk(*cit)) {
				return true;
			}
		}
	}
	return false;
}


// Now the methods of a ConfigPeer
// ================================

/// Try to send a chunk to someone, using a chunk-first strategy.
/// Because of the possibility that the selected chunk can't actually
/// be sent to any of the neighbours, the algorithm loops to until
/// a valid (chunk, peer) combination is found. However, looping
/// should be the exception, not the rule, unless the chunk-selection
/// algorithm is poor or the particlar peer is a disadvantaged one
/// that's unable to provide service to its neighbours.
void ConfigPeer::chunk_first_scheduler() {
	bool keep_trying = false;
	std::vector<PChunk> excluded_chunks;
	do {
		keep_trying = false;
		selected_chunk = select_chunk(&excluded_chunks);
		if (selected_chunk != NULL) {
			selected_peer = select_peer(NULL);
			if (selected_peer == NULL) {
				excluded_chunks.push_back(selected_chunk);  // Give up on that chunk. 
				keep_trying = true;							// Try again.
			}
			else {
				schedule_chunk(selected_chunk, selected_peer);
			}
		}
	} while (keep_trying);
}

/// Try to send a chunk to someone, using a peer-first strategy.
/// Because of the possibility that the selected peer can't actually
/// use any of our available chunks, the algorithm loops to until
/// a valid (peer, chunk) combination is found. However, looping
/// should be the exception, not the rule, unless the peer-selection
/// algorithm is poor or the particlar peer is a disadvantaged one
/// that's unable to provide service to its neighbours.
void ConfigPeer::peer_first_scheduler() {
	bool keep_trying = false;
	std::vector<PLink> excluded_peers;
	do {
		keep_trying = false;  // Default is to give up
		selected_peer = select_peer(&excluded_peers);
		if (selected_peer != NULL) {
			selected_chunk = select_chunk(NULL);
			if (selected_chunk == NULL) {
				excluded_peers.push_back(selected_peer);  // Give up on that peer. 
				keep_trying = true;						  // Try again.
			}
			else {
				schedule_chunk(selected_chunk, selected_peer);
			}
		}
	} while (keep_trying);
}

// Try to send a chunk to someone using the configured chunk
// scheduling strategy.
void ConfigPeer::schedule_new_chunk() {
	if (chunk_trace) printf("T = %lf, peer %d schedule_new_chunk\n", gCurrentTime, id());
	if (!p_upload_busy) {
		switch (sel_order) {
			case ChunkFirst:
				chunk_first_scheduler();
				break;
			case PeerFirst:
				peer_first_scheduler();
				break;
			default:
				fatal("Unknown or unimplemented config-peer strategy");
		} 
	}
}

/// Select a chunk according to the various option settings.  The parameter
/// is a pointer to a vector containing chunks that are to be excluded
/// from the selection (presumably because we already tried one and it
/// didn't work). This can be null to exclude none. 
PChunk ConfigPeer::select_chunk(std::vector<PChunk>* exclusions) {
	if (chunk_sel_type == Best) {
		switch (chunk_filter) {
			case AllChunks:
				return choose_best_chunk(exclusions, NULL, chunk_value);
			case UsefulToSelected:
				return choose_best_chunk(exclusions, useful_to_selected, selected_peer, chunk_value);
			case UsefulToSomeone:
				return choose_best_chunk(exclusions, useful_to_any, 
					std::pair<LinkIterator, LinkIterator>(first_out_link(), last_out_link()), chunk_value);
			default:
				fatal("Unknown or unimplemented chunk-filter type");
		}
		
	}
	else {
		switch (chunk_filter) {
			case AllChunks:
				return choose_weighted_chunk(exclusions, NULL, chunk_value);
			case UsefulToSelected:
				return choose_weighted_chunk(exclusions, useful_to_selected, selected_peer, chunk_value);
			case UsefulToSomeone:
				return choose_weighted_chunk(exclusions, useful_to_any, 
					std::pair<LinkIterator, LinkIterator>(first_out_link(), last_out_link()), chunk_value);
			default:
				fatal("Unknown or unimplemented chunk-filter type");
		}		
	}
	return NULL; // Shut the compiler up
}

/// Select a peer according to the various option settings. The parameter
/// is a pointer to a vector containing peers that are to be excluded
/// from the selection (presumably because we already tried one and it
/// didn't work). This can be null to exclude none. 
PLink ConfigPeer::select_peer(std::vector<PLink>* exclusions) {
	if (peer_sel_type == Best) {
		switch (peer_filter) {
			case AllPeers:
				return choose_best_neighbour(exclusions, not_busy, peer_value);
			case NeedsAny:
				return choose_best_neighbour(exclusions, needs_any_and_not_busy, window(), peer_value);
			case NeedsSelected:
				return choose_best_neighbour(exclusions, needs_chunk_and_not_busy, selected_chunk, peer_value);
			default:
				fatal("Unknown or unimplemented chunk-filter type");
		}
	}
	else {
		switch (peer_filter) {
			case AllPeers:
				return choose_weighted_neighbour(exclusions, not_busy, peer_value);
			case NeedsAny:
				return choose_weighted_neighbour(exclusions, needs_any_and_not_busy, window(), peer_value);
			case NeedsSelected:
				return choose_weighted_neighbour(exclusions, needs_chunk_and_not_busy, selected_chunk, peer_value);
			default:
				fatal("Unknown or unimplemented chunk-filter type");
		}
	}
	return NULL;  // Shuts the compiler up
}

/// Select the chunk-scheduling strategy
/// Sets up the various chunk-scheduling parameters according to the given configuration string.
/// The string is made up of a sequence of tokens as follows:
/// config_string :== selection_order chunk_selector peer_selector
/// selection_order :== ChunkFirst | PeerFirst
/// chunk_selector :== chunk_selection_type chunk_filter chunk_weight
/// peer_selector :== peer_selection_type peer_filter peer_weight
/// peer_selection_type :== BestPeer | WeightedPeer
/// chunk_selection_type :== BestChunk | WeightedChunk
/// chunk_filter :== AllChunks | UsefulToSelected | UsefulToSomeone
/// chunk_weight :== UnitWtChunk | Recentness
/// peer_filter :== AllPeers | NeedsAny | NeedsSelected
/// peer_weight :== UnitWtPeer | UploadBW | DownloadBW | EarliestLatest | BAELp

void ConfigPeer::set_strategy(const char* strategy) {
	std::istringstream in(strategy);
	std::string token;
	while (getline(in, token, ',')) {
	        //std::printf("ConfigPeer:%s\n", token.c_str());
		if (token == "ChunkFirst")
			sel_order = ChunkFirst;
		else if (token == "PeerFirst")
			sel_order = PeerFirst;
		else if (token == "BestPeer")
			peer_sel_type = Best;
		else if (token == "WeightedPeer")
			peer_sel_type = Weighted;
		else if (token == "BestChunk")
			chunk_sel_type = Best;
		else if (token == "WeightedChunk")
			chunk_sel_type = Weighted;
		else if (token == "UsefulToSelected")
			chunk_filter = UsefulToSelected;
		else if (token == "UsefulToSomeone")
			chunk_filter = UsefulToSomeone;
		else if (token == "AllChunks") 
			chunk_filter = AllChunks;
		else if (token == "NeedsAny")
			peer_filter = NeedsAny;
		else if (token == "NeedsSelected")
			peer_filter = NeedsSelected;
		else if (token == "AllPeers")
			peer_filter = AllPeers;
		else if (token == "Recentness")
			chunk_value = &chunk_recentness;
                else if (token == "DLc")
                        chunk_value = &chunk_deadline;
		else if (token == "UploadBW")
			peer_value = &peer_uploadBW;
		else if (token == "DownloadBW")
			peer_value = &peer_downloadBW;
		else if (token == "EarliestLatest")
			peer_value = &peer_latestChunk;
		else if (token == "BAELp")
		        peer_value = &peer_baelp;
		else if (token == "UnitWtChunk")
			chunk_value = NULL;
		else if (token == "UnitWtPeer")
			peer_value = NULL;
		else
			fatal(std::string("Unknown/unimplemented token in ConfigPeer configuration string:")+token);
	}	
}

