/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors: Maurizio Munafo', Marco Mellia, Richard Lobb, Gloria Bogliolo, 
/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Maurizio Munafo', Marco Mellia, Gloria Bogliolo, 
Ana Paula Couto, Richard Lobb
Telecomunication Networks Group
Politecnico di Torino
Torino, Italy
mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
Defines the various subclasses of the Peer class. These are
primarily just different schedulers, i.e., each one defines its own
version of the virtual schedule_chunk method of Peer.
*/

#ifndef _CONFIG_PEER_H
#define _CONFIG_PEER_H

#include <vector>
#include "types.h"
#include "peer.h"
#include "propertymap.h"
#include "globals.h"

//=======================ConfigPeer ==============================
// This is a peer that can be dynamically re-configured over a range of
// chunk scheduling strategies. The strategy to be followed by the peer
// is set by a call to its set_strategy method. The parameter is a string
// which is a sequence of tokens setting various attributes of the chunk
// scheduling strategy. To understand this, first read the documentation in
// algorithms.h of the 'best' and 'weighted' search strategies, which defines
// how an entity of some type is chosen from a set of such entities using
// a filter function to select a subset then a weight function to weight the
// different elements. Roughly speaking, a "best" strategy chooses the
// candidate with the highest weight, while a "weighted" strategy chooses
// one probabilisitically according to the weight. The chunk scheduling
// strategies all use those algorithms to select chunks and peers. The
// different algorithms can be classified as "chunk first" if a chunk is
// selected first and then an appropriate neighbouring peer is chosen, or
// "peer first" if the peer is selected and then an appropriate chunk is
// chosen. The various sub-strategies use different weights and different
// filters, as specified by the configuration string, to filter and weight
// both chunks and peers.
// For more details see the description of the set_strategy method in
// peertypes.cpp.
class ConfigPeer : public Peer
{
public:
	enum SELECT_ORDER {ChunkFirst, PeerFirst, Mixed};
	enum SELECT_TYPE {Best, Weighted};
	enum CHUNK_FILTER {AllChunks, UsefulToSelected, UsefulToSomeone};
	enum PEER_FILTER {AllPeers, NeedsAny, NeedsSelected};

private: // fields
	SELECT_ORDER sel_order;
	SELECT_TYPE chunk_sel_type;
	SELECT_TYPE peer_sel_type;
	CHUNK_FILTER chunk_filter;
	PEER_FILTER peer_filter;
	double (*chunk_value)(PChunk);
	double (*peer_value)(PLink);

	PLink selected_peer;
	PChunk selected_chunk;

private: // methods
	PChunk select_chunk(std::vector<PChunk>* exclusions);
	PLink select_peer(std::vector<PLink>* exclusions);

	/// Try to send a chunk to someone, using a chunk-first strategy.
	void chunk_first_scheduler();

	/// Try to send a chunk to someone, using a peer-first strategy.
	void peer_first_scheduler();

public:
	ConfigPeer(
		Bandwidth upload_bandwidth,		///< The upload_bandwidth of the peer
		Bandwidth download_bandwidth,	///< The download_bandwidth of the peer
		size_t peer_class,				///< The bandwidth class of the peer
		const char* peerconfig = NULL):				///< The bandwidth class of the peer
			Peer(upload_bandwidth, download_bandwidth, peer_class), 
			chunk_value(NULL), 
			peer_value(NULL),
			chunk_filter(AllChunks),
			peer_filter(AllPeers)
	{ 
		if (peerconfig) {
			set_strategy(peerconfig);
		} else {
			//TODO: deprecate this branch
			set_strategy(config.getString("ConfigPeerConfig").c_str());
		}
	}
	virtual void schedule_new_chunk();
	void set_strategy(const char* strategy);  ///< Select the chunk-scheduling strategy
};

#endif //_CONFIG_PEER_H
