/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file event.cpp
	Implementations of the Event structure and the internal Event List.
 */

#include "myassert.h"
#include "event.h"
#include "overlay.h"
#include "peer.h"
#include "chunk.h"
#include "random.h"
#include "globals.h"
#include "videostream.h"

// #include "stats.h"
#include "misc.h"

//! The Event structure -- the type of object placed in the event queue.
//  ====================================================================

struct Event
{
	Time  time;			///< Event time
	size_t id;			///< Event ID
	EventType  type;   ///< Event type
	PPeer   source;		///< Chunk source (unused for CHUNK_INITIATION)
	PPeer   dest;		///< Chunk destination (unused for CHUNK_INITIATION
	PChunk   chunk;		///< Pointer to chunk  

	//! Event constructor
	Event(
		  Time time,		///< The time at which the event should happen
		  size_t event_id,	///< An event identifier, provided by the caller (usually unused)
		  EventType type,	///< The type of event
		  PPeer source,	    ///< The source peer involved with the event (if any)
		  PPeer dest,		///< The destination peer involved with the event (if any)
		  PChunk chunk		///< Pointer to the chunk involved with the event
	) 
		: id(event_id), time(time), type(type), source(source), dest(dest), chunk(chunk) {}
};

// A function object to compare the times of two events
// in an event queue, given the pointers to those events.
struct EventComparator {
	bool operator()(const Event* e1, const Event* e2) {
		if (!gContemporaryDeliveries || e1->time - e2->time != 0) return e1->time > e2->time;	// compare event's timestamp
		return e1->type > e2->type;	// if timestamp is the same, order by event type to anticipate deliveries
	}
};

typedef std::priority_queue<Event*, std::deque<Event*>, EventComparator> EVENT_LIST;

// The EventManager class
// =======================

// The main event list. A static variable of this file rather than a static field of
// the event manager class in order to avoid publishing a whole lot of rubbish on the
// interface. [Sometimes C++ just seems brain-dead!]
static EVENT_LIST event_list;

// Schedule an event of a given type at a given time. 

void EventManager::schedule_event(EventType type, Time time, PPeer source, PPeer dest, 
					PChunk chunk, size_t event_id)
{
	double jittered_time = time + uniform(0, +gTimeJitter, &gSeed);
	event_list.push(new Event(jittered_time, event_id, type, source, dest, chunk));
}

/// An alternative interface to the event-peer_type
void EventManager::schedule_event(EventType type, Time time, Path path, 
					PChunk chunk, size_t event_id)
{
	schedule_event(type, time, path.from, path.to, chunk, event_id);
}

// The main event loop for the program
void EventManager::run_event_loop(POverlay po)
{
	bool simulation_end = false;
	Event *ev;
	while (!event_list.empty() && !simulation_end) {
		ev = event_list.top();
		event_list.pop();
		myassert(ev->time >= gCurrentTime, "Time running backwards!");
		gCurrentTime = ev->time;
		switch (ev->type) {
		case CHUNK_SENT:
			ev->source->chunk_sent(ev->chunk, ev->dest);
			break;
		case READY_TO_SEND:
			ev->source->ready_to_send();
			break;
		case CHUNK_DELIVERED:
			ev->dest->host()->deliver_chunk(ev->chunk, Path(ev->source, ev->dest), ev->id);
			break;
		case CHUNK_INITIATION:
			VideoStream::next_chunk(po);
			break;
		case CHUNK_TIMEOUT:
			ev->dest->chunk_timeout(ev->chunk);
			break;
		case SIMULATION_END:
			simulation_end = true;
			break;
		default:
			fatal("Event does not exist");
		}
		delete ev;
	}
}

