/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
    Defines the GNROverlay, which is an MBWOverlay (i.e., multiple bandwidth
	overlay) in which each peer is connected to a randomly-selected
	fixed-size subset of the other peers.
 */

#include <cmath>
#include <vector>
#include "gnroverlay.h"
#include "random.h"
#include "globals.h"
#include "misc.h"
#include "aopeer.h"
#include "myassert.h"

// Local function to make a random neighbourhood.
void make_random_neighbourhood(POverlay po, size_t degree)
{
	size_t n = po->num_peers();
	for (size_t i = 0; i < n; i++) {
		size_t k = 0;
		while (k < degree){		
			size_t j = (size_t)floor(uniform (0, n, &gSeed));
			if (i != j) {
				k++;
				if (po->find_link (i, j) == NULL){
					po->add_link (i, j);
					po->add_link (j, i);
				}
			}
		}
	}	
}


// The GNROverlay
// -------------

GNROverlay::GNROverlay(size_t num_peers, size_t degree) {
	if(degree > num_peers - 1) {
		fatal("The degree is larger than the total number of peers");
	}
	o_degree = degree;
	set_peers(num_peers);  // See MBWOverlay
	set_edges(degree);
}

// Set up each peer with a random neighbourhood of the
// specified degree.

void GNROverlay::set_edges(size_t degree)
{
	make_random_neighbourhood(this, degree);
}

// Connect peer to 2*degree neighbours.
void GNROverlay::connect_peer(PPeer peer) {
	size_t n = std::min(num_peers()- 2, 2 * o_degree);

	// Get a set of neighbours (1 more than we need in case our own id comes up)
	std::vector<size_t> neighbours = random_ints(n+1, 0, num_peers()-1);

	// Make the connections
	size_t i = 0;
	while (peer->num_links_out() < n) {
		PPeer neighbour = o_peer_set[neighbours[i++]];
		if (peer != neighbour) {
			peer->add_link(neighbour);
			// myassert(find_link(neighbour->id(), peer->id()) == NULL, "GNROverlay::connect: already linked");
			neighbour->add_link(peer);
		}
	}
}

void GNROverlay::select_churners(double percent) {
	size_t num_churners = 0;
	size_t max_peer_num = num_peers() - 1;  // Must exclude source (id 0) for all calcs here
	myassert(percent <= 100, "Overlay::select_churners: can't have more than 100% churners!");
	size_t n = size_t(max_peer_num * percent / 100);  // Num churners
	std::vector<size_t> selected = random_ints(n, 1, max_peer_num);
	for (size_t i = 0; i < n; i++) {
		((AOPeer*) o_peer_set[selected[i]])->set_churn(true);
	}
}
