/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	A file containing the declarations of the Host class. A Host is a peer's
	access point to a network -- all peer-to-peer chunk communication is sent
	via hosts. There may be multiple peers sharing a single host, but not
	vice-versa.
 */

#ifndef _HOST_H_
#define _HOST_H_

#include <vector>
#include <string>
#include "types.h"

/// HostType is an enumeration of sub-classes of Host, which implement
/// different network-level chunk routing algorithms.
enum HostType { OLD_SIM, TWEAKED_OLD_SIM, PRIORITY_SHARING, BAD_HOST_TYPE };

/// An array of strings corresponding one-for-one with HostType. Used by
/// the main program to map from the HostType entry in config.txt to the
/// internal HostType.
extern std::string gHostTypeNames[];

class Host
{
protected:
	HostID h_id;				///< The ID of this host. Defined by the network, if used, else 0.
	PRouter h_router;			///< Pointer to the router to which this host is connected
	Bandwidth h_uploadBW;		///< The host's upload bandwidth (MHz)
	Bandwidth h_downloadBW;		///< The host's download bandwidth (MHz)
	Bandwidth h_download_avail; ///< The host's currently-available download bandwidth (MHz)
	Latency h_upload_latency;	///< The latency from host to router (above and beyond the bandwidth-induced delay)
	Latency h_download_latency;	///< The latency from router to host (above and beyond the bandwidth-induced delay)

	static HostType h_host_type;  ///< Defines what subclass of Host is generated by the host factory.

public:
	/// Construct a host with given bandwidth parameters. 
	/// Since Host is an abstract class, this will always be
	/// called via one of its subclasses.
	Host(Bandwidth uploadBW, Bandwidth downloadBW,
		Latency upload_latency, Latency download_latency);

	/// Define the type of host to be generated by a call to the host factory method,
	/// create_host.
	static void set_host_type(HostType type) { h_host_type = type; }

	/// Called by the network (if one is used) to set an ID for this host,
	/// which is currently equal to the network's index to this host in its host table.
	/// Will be zero if no network is used, i.e., if communication passes directly
	/// from the source host to the destination host.
	void set_id(HostID id);

	/// Connect this host to a router. Used only if a network is also being used,
	/// since the role of a router is just as an access point to the network.
	void set_router(PRouter router);

	inline PRouter router() { return h_router; }

	/// The bandwidth from this host to its router (whether explicitly defined
	/// or not). MHz
	inline Bandwidth upload_bandwidth() { return h_uploadBW; }

	/// The bandwidth from the router (whether explicitly defined
	/// or not) to this host. MHz
	inline Bandwidth download_bandwidth() { return h_downloadBW; }

	/// The currently-available download bandwidth. MHz.
	/// Used by chunk-schedulers to decide whether or not to send a chunk
	/// to peers on this host. Note: the existence of this method is
	/// somewhat controversial, as it's not something generally available
	/// on a real host over a short time scale.
	inline Bandwidth download_BW_avail() { return h_download_avail; }

	/// The additional latency from this host to its router (whether explicitly
	/// defined or not) on top of the bandwidth-determined transmission delay.
	inline Latency upload_latency() { return h_upload_latency; }

	/// The additional latency from the router (whether explicitly
	/// defined or not) to this host, on top of the bandwidth-determined
	/// transmission delay.
	inline Latency download_latency() { return h_download_latency; }

	/// Send the given chunk along the given path. This method is expected
	/// to schedule any events associated with the source end of the path
	/// and then to call the destination host's chunk_scheduled method to
	/// schedule events associated with an incoming chunk from the network.
	virtual void send_chunk(PChunk chunk, Path path) = 0;

	/// Notify the target host that a chunk is on its way so it can 
	/// schedule events associated with the download end of the path, i.e.,
	/// the potential bottleneck between the destination router and the
	/// destination host. This method is called by the source host or
	/// by the network to schedule whatever callback events the desintation
	/// host may wish to receive as the chunk passes through it. At very
	/// least it must schedule an event responsible for passing the chunk on
	/// to the target peer an appropriate time later.
	virtual void chunk_scheduled(PChunk chunk, Path path) = 0;

	/// Deliver the given chunk to the target peer (assumed to be attached
	/// to this host). Called via an CHUNK_DELIVERED event that was
	/// scheduled by a prior call to chunk_scheduled.
	virtual void deliver_chunk(PChunk chunk, Path path, size_t event_id = 0);

	/// The host factory. Constructs and returns a pointer to a host of
	/// the type specified by the most recent call to set_host_type.
	static PHost create_host(Bandwidth uploadBW, Bandwidth downloadBW,
		 Latency upload_latency = 0, Latency download_latency = 0);
};

#endif // _HOST_H_
