/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file main.cpp
	The main driver function for the simulator. Processes command-line
	arguments, builds overlays, initiates and runs a simulation, prints results.
 */

#include <cstdio>
#include <cmath>
#include <string>
#include <fstream>
#include <cstdlib>
#include <algorithm>
#include <iostream>
#include "overlay.h"
#include "link.h"
#include "peer.h"
#include "misc.h"
#include "random.h"
#include "event.h"
#include "globals.h"
#include "propertymap.h"
#include "asnetwork.h"
#include "videostream.h"
#include "gnroverlay.h"

POverlay po = NULL;  // Pointer to the Overlay object

// Find the number of outlinks of peers in the specified ID range


// Polling function, called via the VideoStream, to print the link status at regular
// intervals and clear the statistics at a specified time.
void report_status(size_t chunk_number) {
	static FILE* fDegreeStatus = check_open("DegreeOverTimeFile");
	std::map<size_t, size_t> in_link_counts;
	std::map<size_t, size_t> out_link_counts;
	std::map<size_t, size_t> num_in_class;
	for (PeerIterator it = po->first_peer(); it != po->last_peer(); ++it) {
		size_t peer_class = (*it)->get_class();
		num_in_class[peer_class]++;
		in_link_counts[peer_class] += (*it)->num_links_in();
		out_link_counts[peer_class] += (*it)->num_links_out();
	}

	fprintf(fDegreeStatus,"%d %d", chunk_number, po->num_arcs());
	for (size_t i = 0; i < num_in_class.size(); ++i) {
		fprintf(fDegreeStatus," %d %.1f %.1f", num_in_class[i],
			in_link_counts[i]/(float) num_in_class[i], out_link_counts[i]/(float) num_in_class[i]);
	}
	fprintf(fDegreeStatus,"\n");
	printf("Chunk %d\n", chunk_number);
}

void reset_stats(size_t chunk_number) {
	printf("Resetting stats\n");
	po->restart_peer_stats();
}

/****************************************************/
/* Main Function                                    */
/****************************************************/		

int main (int argc, char *argv[])
{
	config.load("config.txt");
	config.process_command_line_args(argc, argv);
	initialize_globals();
	FILE* dump_file = check_open("ConfigDumpFile");
	config.dump(dump_file);
	long runseed = 0;
	if (config.has_key("DiffusionSeed")) {
		runseed = config.getLong("DiffusionSeed");
	}
	
	size_t num_peers = config.getSizeT("NumPeers");
	size_t degree = config.getSizeT("Degree");
	size_t num_chunks = config.getSizeT("NumChunks");

	// Determine if full chunk-arrival recording is required

	PeerStats::set_chunk_arrival_recording(config.has_key("ChunkArrivalsFile") || config.has_key("ChunkDelaysFile"));
	
	// Set up a network object

    std::string* p;
    std::string net_type_name = config.getString("NetworkType");
    if(net_type_name != "NONE")
    {
        std::string* p = std::find(&gNetworkTypeNames[0], &gNetworkTypeNames[ILLEGAL], net_type_name);
        NetworkType network_type = (NetworkType) (p - gNetworkTypeNames);
        if (network_type == BAD_NET) {
                fatal("Bad argument: unknown network type:"+net_type_name);
        }
        std::cout << "Creating Network of type " << net_type_name << std::endl;
        Network::set_network_type(network_type);
        Network::make_network();
    }
    else
    {
        printf("The Network is not being created\n");
    }
	

	// Lookup the overlay type and set it in the Overlay class.

	std::string type_name = config.getString("OverlayType");
	p = std::find(&gOverlayTypeNames[0], &gOverlayTypeNames[ILLEGAL], type_name);
	OverlayType overlay_type = (OverlayType) (p - gOverlayTypeNames);
	if (overlay_type == ILLEGAL) {
		fatal("Bad argument: unknown overlay type:"+type_name);
	}
	Overlay::set_overlay_type(overlay_type);

	// Lookup the peer type and set it in the Peer class.

	std::string peer_type_name = config.getString("PeerType");
	p = std::find(&gPeerTypeNames[0], &gPeerTypeNames[BAD], peer_type_name);
	PeerType peer_type = (PeerType) (p - gPeerTypeNames);
	if (peer_type == BAD) {
		fatal("Bad argument: unknown peer type:"+peer_type_name);
	}
	Peer::set_peer_type(peer_type);

	// Lookup the host type and set it in the Host class.

	std::string host = config.getString("HostType");
	p = std::find(&gHostTypeNames[0], &gHostTypeNames[BAD_HOST_TYPE], host);
	HostType host_type = (HostType) (p - gHostTypeNames);
	if (host_type == BAD_HOST_TYPE) {
		fatal("Bad argument: unknown host type:"+host);
	}
	Host::set_host_type(host_type);

	printf("Peer type %s\n", gPeerTypeNames[peer_type].c_str());
	printf("creating overlay\n");

	// Set up the overlay. This uses the first argument gSeed.

	po = Overlay::make_overlay(num_peers, degree);
	
	FILE* fp_link = check_open("LinksFile");
	FILE* fp_degree = check_open("DegreeFile");
	po->print_overlay(fp_degree, fp_link);
	printf ("Overlay: %s, total number of peers %d, mean degree %.2f, total number of arcs %d\n",
		type_name.c_str(), po->num_peers(), po->num_arcs() / (float) po->num_peers(), po->num_arcs());	
	fclose(fp_link);
	fclose(fp_degree);

	printf("Source Node ID %d\n", po->source());
	printf("Total Number of chunks %d\n", num_chunks);

	// Set up a callback from the video stream for regular overlay status reports
	if (config.has_key("StatusReportInterval")) {
		VideoStream::add_callback(&report_status, config.getInt("StatusReportInterval"), true); 
	}

	// Set up a callback to clear the stats records at a specified chunk_number (if given)
	if (config.has_key("StartStats")) {
		size_t firstID = config.getInt("StartStats");
		if (firstID != 0) {
			VideoStream::add_callback(&reset_stats, firstID);
			PeerStats::set_first_chunk_to_record(firstID);	//TODO: this seems to be duplicated in reset_stats
		}
	}

	// Set up the overlay and peers to stop gathering statistics as configured
	size_t lastID;
	if (config.has_key("StopStats") && (lastID = config.getInt("StopStats"))) {
		po->set_last_chunk_to_record(lastID);
		PeerStats::set_last_chunk_to_record(lastID);	//TODO: do we need to stop any other stats?
	} else {
		po->set_last_chunk_to_record(num_chunks-1);
		PeerStats::set_last_chunk_to_record(num_chunks-1);	//TODO: do we need to stop any other stats?
	}

	// Set up churning if config file set up for it
	if (config.has_key("ChurnPercent")) {
		if (config.getString("OverlayType") != "GNR" ||
			config.getString("PeerType") != "AO_PEER") {
				fatal("Churning requires GNROverlay and AOPeer");
		}
		((GNROverlay*) po)->select_churners(config.getDouble("ChurnPercent"));
	}

	// Now run a chunk diffusion, using the second gSeed (if given).

	printf ("Chunk diffusion begins ... \n");

	if (runseed != 0) {
		gSeed = runseed;
	}

	po->init_chunk_diffusion(num_chunks);
	EventManager::run_event_loop(po);

	// Dump all statistics to the various stats files

	printf("Printing results...\n");
	po->print_statistics();

	if (config.has_key("LinksAtEndFile")) {
		fp_link = check_open("LinksAtEndFile");
		fp_degree = check_open("DegreeAtEndFile");
		po->print_overlay(fp_degree, fp_link);
		fclose(fp_link);
		fclose(fp_degree);
	}

	return EXIT_SUCCESS;
}
