/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file 
    Implementation of the OldSimHost subclass of the Host class.
	This is a host that mimics the behaviour of the old simulator (which
	lacked a network layer).
 */

#include "myassert.h"
#include "oldsimhost.h"
#include "globals.h"
#include "path.h"
#include "event.h"
#include "router.h"
#include "peer.h"
#include "misc.h"

/// An OldSimHost is intended to mimic the behaviour of the old simulator.
/// The total chunk delivery time is computed, based on bandwidths and
/// latencies provided by the routers and the network. Events are then
/// scheduled for that time to notify the sender and the receiver.

OldSimHost::OldSimHost(Bandwidth uploadBW, Bandwidth downloadBW,
		   Latency upload_latency=0, Latency download_latency=0)
		   : Host(uploadBW, downloadBW, upload_latency, download_latency) {}

void OldSimHost::send_chunk(PChunk chunk, Path path) {
	double effective_bandwidth = std::min(upload_bandwidth(), path.to->download_bandwidth());
	double transmission_delay = chunk->size() / effective_bandwidth;
	double latency = upload_latency() + path.to->host()->download_latency();
	if (router() != NULL) {
		latency += router()->network_latency(path);
	}
	double arrival_time = gCurrentTime + transmission_delay + latency;
	EventManager::schedule_event(CHUNK_SENT, arrival_time, path.from, path.to, chunk);

	if (uniform(0, 1, &gSeed) < gProbChunkError) {
		if (gChunkTimeoutInterval<0) {
			EventManager::schedule_event(CHUNK_TIMEOUT, arrival_time, path.from, path.to, chunk);
		} else {
			EventManager::schedule_event(CHUNK_TIMEOUT, gCurrentTime + gChunkTimeoutInterval, path.from, path.to, chunk);
		}
	} else {
		EventManager::schedule_event(CHUNK_DELIVERED, arrival_time, path.from, path.to, chunk);
	}

	if (router() != NULL) {
		router()->sending_chunk(chunk, path);  /// Notify router of incoming chunk
	}

	path.to->host()->chunk_scheduled(chunk, path);
}

void OldSimHost::deliver_chunk(PChunk chunk, Path path, size_t event_id)
{
	myassert(is_infinite(download_bandwidth()) || 
		h_download_avail == 0, "OldSimHost: unexpected chunk arrival");
	h_download_avail = download_bandwidth();  /// Mark download channel as available again
	path.to->receive_chunk(chunk, path.from);
}

/// In the old sim, the sending of a chunk is simultaneous with its receipt,
/// so most of the work has already been done by send-chunk.
/// All that's necessary is to mark the download channel as busy,
/// i.e., no BW available, for the case of a download-limited host.
void OldSimHost::chunk_scheduled(PChunk chunk, Path path) {
	myassert(h_download_avail != 0, "OldSimHost: multiple chunks scheduled?!");
	if (!is_infinite(download_bandwidth())) {
		h_download_avail = 0;
	}
}
