/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

#include <algorithm>
#include <cmath>
#include "overlay.h"
#include "peer.h"
#include "link.h"
#include "event.h"
#include "random.h"
#include "misc.h"
#include "globals.h"
#include "chunk.h"
#include "videostream.h"
#include "myassert.h"

/*! @file
	Implementations of the various Overlay methods. This excludes those
	methods declared inline in overlay.h and also all statistics-related
	methods, which are in stats.cpp.
 */

/// The overlay names in the following line *must* match one-for-one the overlay types
/// declared in the overlay type enumeration (see overlay.h).
std::string gOverlayTypeNames[] = {"FULL", "GNR", "GNP", "GNKREG", "TREE", "BWO", "PREBUILT", "TEST", "AS_AWARE"};

Overlay::Overlay(): o_src(0), o_num_arcs(0), o_stats_start(0) { }

OverlayType Overlay::o_type = ILLEGAL;

// Add a new peer to the overlay
PeerID Overlay::add_peer(PPeer peer)
{
	PeerID id = o_peer_set.size();
	o_peer_set.push_back(peer);
	peer->set_id(id);
	peer->set_overlay(this);
	return id;
}

/// Notification from a peer that it just added a link
void Overlay::link_added(PPeer source, PPeer dest) {
	o_num_arcs++;
}

/// Notification from a peer that it just deleted a link
void Overlay::link_removed(PPeer source, PPeer dest){
	o_num_arcs--;
}

// Add a new link from->to into the overlay. 
void Overlay::add_link (PeerID from, PeerID to)
{
	peer(from)->add_link(peer(to));
}

// Remove the given link from the overlay.
void Overlay::remove_link (PPeer from, PLink p_link)
{
	from->delete_link(p_link);
}

// Returns a pointer to the link from peer 'from' to peer 'to' or NULL if no such link.
PLink Overlay::find_link (PeerID from, PeerID to)
{
	PPeer source = peer(from);
	PPeer dest = peer(to);
	for (LinkIterator it = source->first_out_link(); it != source->last_out_link(); ++it) {
		if ((*it)->to() == dest) {
			return *it;
		}
	}
	return (NULL);
}

// Reset overlay to delete simulation data stored at peers
void Overlay::clear()
{
	for (PeerIterator it = first_peer(); it != last_peer(); ++it) {
		(*it)->clear(); 
	}  
}

// Restart collection of peer statistics, but without affecting
// per link throughput.
void Overlay::restart_peer_stats()
{
	o_stats_start = VideoStream::get_next_chunk_num();
	PeerStats::set_first_chunk_to_record(o_stats_start);
	for (PeerIterator it = first_peer(); it != last_peer(); ++it) {
		(*it)->reset_peer_stats(); 
	} 
}

void Overlay::set_last_chunk_to_record(size_t lastID)
{
	o_stats_stop = lastID;
	PeerStats::set_last_chunk_to_record(o_stats_stop);
}

void Overlay::init_chunk_diffusion(size_t num_chunks)
{
	clear();
	VideoStream::clear();
	VideoStream::set_num_chunks(num_chunks);
	EventManager::schedule_event(CHUNK_INITIATION, 0, NULL, NULL, NULL);
}

// Print the overlay description to the Peers.out, Links.out and Degree.out files.
void Overlay::print_overlay(FILE* fp_degree, FILE* fp_link) {
	float mean_degree = num_arcs() / (float) num_peers();

	const char* type_name = gOverlayTypeNames[o_type].c_str();
	fprintf(fp_link,"Overlay: %s, Mean degree %.2f, Total number of arcs %d\n", type_name, mean_degree, num_arcs());

	for(PeerIterator pit = first_peer(); pit != last_peer(); ++pit) {
		PPeer from = *pit;
		double upbw = from->upload_bandwidth();
		double downbw = from->download_bandwidth();

		fprintf(fp_degree,"Peer %d\t%d\t%f\t%f\n", from->id(), from->num_links_out(),
			upbw, downbw); 
		for (LinkIterator it = from->first_out_link(); it != from->last_out_link(); ++it) {
			fprintf(fp_link,"Source %d \tDestination %d\n", from->id(), (*it)->to()->id());
		}
	}

	fprintf(fp_degree,"Overlay: %s, Mean degree %.2f, Total number of arcs %d\n", type_name, mean_degree, num_arcs());

}

void Overlay::connect_peer(PPeer peer)
{
	myassert(true, "Overlay::connect_peer: unimplemented");
}



