/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Maurizio Munafo', Marco Mellia, Gloria Bogliolo, 
		  Ana Paula Couto, Richard Lobb,
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
		  mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file overlay.h
	Defines the Overlay class. This is the class used to represent the peers
	in the network and the links between them. Chunks are diffused through
	a overlay, and statistics are recorded within the overlay for subsequent
	printing with its <i>print_statistics</i> method. This is an abstract
	class as it does not contain code to set up peers and edges. That is
	done in subclasses; see <i>overlaytypes.h</i> for some examples.
 */
#ifndef _OVERLAY_
#define _OVERLAY_

#include <cstdio>
#include <vector>
#include <string>

#include "types.h"

enum OverlayType { FULL, GNR, GNP, GNKREG, TREE, BWO, PREBUILT, TESTING_OVERLAY, AS_AWARE, ILLEGAL };

extern std::string gOverlayTypeNames[];  ///< String representations of overlay types

/*! The Overlay class. */
class Overlay
{
protected:
	PeerID o_src;					///< The source peer for the video stream
	static OverlayType o_type;		///< The required subclass, for use by the factory method
	size_t o_num_arcs;				///< Total number of directed arcs in the overlay
	size_t o_stats_start;			///< The chunk number at which stats gathering begins
	size_t o_stats_stop;			///< The chunk number at which stats gathering stops
	std::vector<PPeer> o_peer_set;	///< The set of peers. Subclasses build this.

public:
	Overlay();			///< The constructor, which just initialises the fields to zero

	/// Select the type of overlay to be generated by calls to the factory method make_overlay
	static void set_overlay_type(OverlayType type) { o_type = type; }

	// Methods that operate on a overlay
	// ===============================
	void clear();				///< Reset overlay to delete chunk diffusion info

	/// Clear overlay, set up event queue for the given number of chunks, initialise
	/// chunk list.
	void init_chunk_diffusion(size_t num_chunks); 

	// Restart collection of peer statistics, but without affecting
	// per link throughput.
	void restart_peer_stats();

	// set when to end collection of peer statistics, 
	void set_last_chunk_to_record(size_t lastID);

	/// Add a peer to the overlay. Return the PeerID of the newly
	/// added peer. The peer will be unconnected to other peers in
	/// the overlay. This version of the call, without
	/// a host, is for use when no network layer is involved. Otherwise the version
	/// with a host parameter should be used.
	PeerID add_peer(PPeer peer);

	/// Add a directed link to the overlay between the two peers
	void add_link (PeerID from, PeerID to);

	/// Find the link from one given peer to another.
	/// @return a pointer to the required link or NULL if no such link exists.
	PLink find_link (PeerID from, PeerID to);

	/// The cartesian distance between peers i and j
	double distance( PeerID i, PeerID j);	

	/// Remove the specified link from the overlay.
	void remove_link (PPeer from, PLink p_link);

	/// Notification from a peer that it just added a link
	void link_added(PPeer source, PPeer dest);

	/// Notification from a peer that it just deleted a link
	void link_removed(PPeer source, PPeer dest);

	/// Connect the given peer into the overlay in a manner that approximates
	/// the original overlay-building method as well as possible. This method is
	/// used to connect in a new peer after the original overlay has been built.
	virtual void connect_peer(PPeer peer);

	// Methods to access fields of a overlay
	// ===================================
	/// The number of peers in the overlay (including the source)
	inline size_t	num_peers()		{ return o_peer_set.size(); }

	/// The number of arcs in the overlay
	inline size_t	num_arcs()		{ return o_num_arcs; }

	/// An iterator over the set of peers
	inline PeerIterator first_peer() { return o_peer_set.begin(); }
	inline PeerIterator last_peer()  { return o_peer_set.end(); }

	/// A pointer to the peer with given ID
	inline PPeer	peer(PeerID id)	{ return o_peer_set[id]; }

	/// The ID of the chunk-source peer of the overlay
	inline PeerID	source()		{ return o_src; }

	/// A pointer to the chunk-source peer of the overlay
	inline PPeer	source_peer()	{ return o_peer_set[o_src]; }

	// Method to print the overlay to two given files (why 2? History!).
	void print_overlay(FILE* degreeFile, FILE* linkFile);

	// overlay-construction factory method
	// =================================
	static POverlay make_overlay(size_t num_peers, size_t degree);

	/// Build a global link list for use by all peers 

	// Statistics 
	// ==========
	// The methods here are hooks for app-dependent statistics gathering.
	// Hence these methods are not defined in overlay.cpp but elsewhere
	// (probably stats.cpp).

	/// A hook to an app-dependent method to record that a particular link
	/// has just had a peer sent on it.
	void stats_chunk_sent(PPeer from, PLink link, PChunk chunk);

	/// A hook to an app-dependent method to print all accumulated statistics.
	virtual void print_statistics();
};

#endif


