/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Maurizio Munafo', Marco Mellia, Gloria Bogliolo, Ana Paula Couto,
		  Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file peer.h
	Implementations of the Peer methods declared in peer.h, except for
	those already declared inline.
 */

#include "myassert.h"
#include <cmath>
#include <limits>
#include "link.h"
#include "peer.h"
#include "random.h"
#include "misc.h"
#include "event.h"
#include "host.h"
#include "globals.h"
#include "chunk.h"
#include "overlay.h"
#include "videostream.h"
#include "algorithms.h"
#include "propertymap.h"

/// The strings in this array must exactly match, one for one, the type names defined
/// in the PeerType enum (see peer.h).
std::string gPeerTypeNames[] = {
	"LATEST_RANDOM", "LATEST_BA", "CONFIG_PEER", "AO_PEER", "ANA", "TEST", "AS_AW_PEER",  "ROX",
};

PeerType Peer::p_peer_type = BAD;

/// Add a link to a new neighbour.
void Peer::add_link(PPeer dest) {
	myassert(p_po->find_link(id(), dest->id()) == NULL, "Peer::add_link: replicated link");
	p_out_links.push_back(new Link(this, dest));
	dest->add_feeder(this);
	p_po->link_added(this, dest);
}

/// Delete the specified link from the neighbourhood
LinkIterator Peer::delete_link(LinkIterator it) {
	PPeer dest = (*it)->to();
	stats()->record_link_deletion(dest);
	myassert(dest != this, "Peer::delete_link: cyclic loop encountered");
	dest->remove_feeder(this);
	p_po->link_removed(this, dest);
	delete (*it);
	return p_out_links.erase(it);
}

/// Delete the specified link from the neighbourhood
void Peer::delete_link(PLink link) {
	LinkIterator it = std::find(first_out_link(), last_out_link(), link);
	myassert(it != last_out_link(), "Peer: Attempt to delete a non-existent link");
	delete_link(it);
}


/// Delete the link to the specified peer from the neighbourhood
void Peer::delete_link(PPeer p) {
	for (LinkIterator it = first_out_link(); it!=last_out_link(); it++){
		if ((*it)->to()==p) {
			delete_link(it);
			return;			
		}
	}
	fatal("Peer: Attempt to delete a non-existent link");
}

PeerSet Peer::neighbour_set() {
	PeerSet result;
	for (LinkIterator it = first_out_link(); it != last_out_link(); ++it) {
		result.insert((*it)->to());
	}
	return result;
}

/// Add to our feeder set the given feeder (source) peer.
void Peer::add_feeder(PPeer source) {
	p_feeders.insert(source);
}

/// Remove a given peer from our feeder set
void Peer::remove_feeder(PPeer source) {
	PeerSet::iterator it = p_feeders.find(source);
	myassert(it != p_feeders.end(), "Peer::remove_feeder: non-existent feeder");
	p_feeders.erase(it);
}

/// Clear the peer state (window, stats etc)
void Peer::clear() {
	window()->clear();
	p_upload_busy = false;
	p_stats.clear();
	for (LinkIterator it = first_out_link(); it != last_out_link(); it++) {
		(*it)->clear();
	}
}

/// Reset the peer statistics gathering but not the link stats (which are needed, e.g.
/// for adaptive overlay configuration).
void Peer::reset_peer_stats() {
	p_stats.clear();
}

/// Called by the peer to send a chunk to another peer via this peer's associated host.
void Peer::send_chunk(PChunk chunk, Path path)
{
	if (chunk_trace) printf("T = %lf, peer %d sending chunk %d to peer %d, %ld\n", gCurrentTime, id(), chunk->id(), path.to->id(), gSeed);
	p_phost->send_chunk(chunk, path);
}

// Event Callback: Notify the peer of the arrival of a chunk.
// The stats are updated accordingly, the chunk is inserted into the window
// and, if the peer is not currently busy, a new chunk upload is (possibly)
// scheduled.
void Peer::receive_chunk(PChunk pchunk, PPeer from)
{
	if (window()->has_chunk(pchunk)) {
		//printf("T = %lf, peer %d has received duplicate chunk %d\n", gCurrentTime, id(), pchunk->id());
		return;
	}

	p_tot_chunks_recd++;
        PChunk pt = new ChunkInstance(pchunk);
	window()->insert(pt, stats());
	stats()->record_arrival(pt);
	if (chunk_trace && from != NULL) {
		printf("T = %lf, peer %d has received chunk %d from peer %d, %ld\n", gCurrentTime, id(), pchunk->id(), from->id(), gSeed);
	}
	try_to_send();
}

// Event Callback: Notify the peer that a chunk it was sending to the given
// dest has now been sent, allowing the peer to schedule another upload.
void Peer::chunk_sent(PChunk chunk, PPeer dest)
{
	if (chunk_trace) printf("T = %lf, peer %d has finished sending chunk %d to peer %d, %ld\n", gCurrentTime, id(), chunk->id(), dest->id(), gSeed);
	p_upload_busy = false;
	try_to_send();
}

void Peer::try_to_send()
{
	if (!p_upload_busy) {
		if (gContemporaryDeliveries) {
			ready_to_send();
		} else {
			EventManager::schedule_event(READY_TO_SEND, gCurrentTime, this , NULL , NULL);
		}
	}
}

// Event Callback: Notify the peer that chunks it was sending 
//  have now been sent and chunks in arrival have already arrived, allowing the peer to schedule another upload.
void Peer::ready_to_send()
{
	if (chunk_trace) printf("T = %lf, peer %d is ready to send, %ld\n", gCurrentTime, id(), gSeed);
	schedule_new_chunk();
	//p_upload_busy = true;
}

// Handle the chunk timeout event. This event is generated at network level
// whenever a chunk can't be delivered because the destination router->host
// link is saturated.
void Peer::chunk_timeout(PChunk chunk)
{
	if (chunk_trace) printf("T = %lf, peer %d has received a timeout notice on chunk %d, %ld\n", gCurrentTime, id(), chunk->id(), gSeed);
	window()->clear_pending(chunk);  // We didn't get it after all
}



//********************************************************************
//
// Chunk SCHEDULING FUNCTIONS
//
//********************************************************************

// Schedule the given chunk for transmission on the given link
void Peer::schedule_chunk(PChunk chunk, PLink current_link) {
	ChunkDeadline dls = chunk->dl();
        chunk->set_dl(dls + gDLIncrement);
        PPeer dest = current_link->to();
	dest->notify_pending(chunk);
	p_upload_busy = true;
	stats()->record_send(chunk);
	current_link->stats()->inc_throughput();
	send_chunk(chunk, Path(this, dest));  // Event scheduling takes place down there
}

// Selector and weighting functions for peer and chunk selection
// =============================================================
// For use with choose_best_neighbour and choose_weighted_neighbour.

/// Selector for a neighbour that doesn't have the selected chunk (recorded in the global
/// variable g_selected_chunk) and isn't busy

bool needs_chunk_and_not_busy(const PLink link, PChunk pchunk) {
	PPeer dest = link->to();
	if (chunk_trace) printf("\t\t\t peer %d chunk %d: needs?%d busy?%d\n", dest->id(), pchunk->id(), dest->needs_chunk(pchunk), dest->download_busy() );
	return dest->needs_chunk(pchunk) && !dest->download_busy();
}

/// Weighting method to weight a neighbour link according to the destination's
/// upload bandwidth.
double peer_uploadBW(const PLink link) {
	return link->to()->upload_bandwidth();
}

/// Weighting method to weight a neighbour link according to the destination's
/// download bandwidth.
double peer_downloadBW(const PLink link) {
	return link->to()->download_bandwidth();
}

/// Weighting method to weight a neighbour link according to the age of the destination's
/// latest chunk.
double peer_latestChunk(const PLink link) {
	return VideoStream::get_next_chunk_num()-link->to()->latest_w_pending();
}

/// Weighting function for the BAELp algorithm.
double peer_baelp(const PLink link) {
 double w = config.has_key("WeightBAELp") ? config.getDouble("WeightBAELp") : 3.0;
 
 // if w==0    => ELp
 // if w>>1    => BAhELp
 // if w<<1    => ELhBAp
 // if w->inf  => BA
 
 // Saturation condition: download bw dest must be >= upload bw src
 if (link->to()->download_bandwidth() >= link->from()->upload_bandwidth()) {
   return w * (link->to()->upload_bandwidth() / link->from()->upload_bandwidth() ) +
     VideoStream::get_next_chunk_num()-link->to()->latest_w_pending();
 }
 
 else {
   //return 0.01; // uncomment this to use the saturation condition
   return w * (link->to()->upload_bandwidth() / link->from()->upload_bandwidth() ) +
     VideoStream::get_next_chunk_num()-link->to()->latest_w_pending();
 }
}

/// Weighting function that rewards the most recent chunks.
double chunk_recentness(PChunk chunk) {
	//return chunk->id();
	return chunk->initiation_time();
}

/// Weighting function that rewards the chunk with the lowest deadline value
double chunk_deadline(PChunk chunk) {
    return -chunk->dl();
}
