#include "myassert.h"
#include <limits>
#include <cmath>
#include <algorithm>

#include "prioritybasedhost.h"
#include "globals.h"
#include "path.h"
#include "event.h"
#include "router.h"
#include "misc.h"

const double DELTA = 0.00000001;

/// A PriorityBasedHost allows multiple chunks from lower-speed sources
/// to be downloaded by the host at the same time, but in a prioritised manner.
/// Available download bandwidth
/// is shared among the chunks in a first-come-first-served manner, i.e., the
/// first chunk gets as much bandwidth as it can use (determined by the source's
/// upload bandwidth), the second gets as much of what's left over as it can use,
/// and so on. However, if an attempt is made to send a chunk to a destination
/// host that is saturated, i.e., has zero free bandwidth, the chunk is treated
/// like an error chunk, and is delivered to its destination as a "chunk_timeout"
/// call.

size_t Download::next_id = 0;  // Static field of Download struct. Inc'd on each new construction.

static bool download_comparator(const Download& dl1, const Download& dl2) {
	return dl1.end_time < dl2.end_time;
}

PriorityBasedHost::PriorityBasedHost(Bandwidth uploadBW, Bandwidth downloadBW,
		   Latency upload_latency=0, Latency download_latency=0)
		   : Host(uploadBW, downloadBW, upload_latency, download_latency) {}

void PriorityBasedHost::send_chunk(PChunk chunk, Path path) {
	double source_send_time = chunk->size() / upload_bandwidth();
	double source_delay = source_send_time + upload_latency();
	double source_delivery_time = gCurrentTime + source_delay;
	EventManager::schedule_event(CHUNK_SENT, source_delivery_time, path, chunk);
	path.to->host()->chunk_scheduled(chunk, path);
	if (router() != NULL) {
		router()->sending_chunk(chunk, path);  /// Notify router of incoming chunk
	}
}

typedef std::list<Download>::iterator PendingIter;

/// Compute the completion time of a new chunk download request, given that
/// all downloads in the given pending list (ordered in increasing order of end_time)
/// take priority for bandwidth utilisation over this new request.
Time chunk_arrival_time(Bandwidth available,  std::list<Download> pending,
						Bandwidth sourceBW, Bandwidth destBW) {

	Time t = gCurrentTime;
	Time end_time = 0;

	PendingIter it = pending.begin(), last = pending.end();
        double bits_remaining = it->chunk->size();
	while (it != last && bits_remaining > 0) {
		Time deltaT = it->end_time - t;			// Time until next chunk finishes
		double rate = std::min(available, sourceBW);
		double bits_this_time = rate * deltaT;  // How much we get done by then
		if (bits_this_time >= bits_remaining) {	// Do we finish before then?
			end_time = t + bits_remaining / rate;   // Yes, compute just when
			bits_remaining = 0;
		}
		else {
			bits_remaining -= bits_this_time;
			t = it->end_time;
			available = std::min(destBW, available + it->sourceBW);
			++it;
		}
	}
	if (bits_remaining > 0) {
		myassert((available == std::numeric_limits<double>::infinity() &&
			destBW == std::numeric_limits<double>::infinity()) || (fabs(available - destBW) < DELTA),
			"Priority sharing algorithm failure #1");
		end_time = t + bits_remaining/std::min(sourceBW, destBW);
	}
	myassert(end_time >= gCurrentTime, "Priority sharing algorithm failure #2");
	return end_time;
}

/// Compute the CHUNK_DELIVERED event associated with downloading the given chunk to
/// this host, given current download traffic and the source bandwidth.
/// Note that all latencies are accumulated and added on at the end of the bandwidth-based
/// calculation.
void PriorityBasedHost::chunk_scheduled(PChunk chunk, Path path) {
	if (h_download_avail <= 0) {
		EventManager::schedule_event(CHUNK_TIMEOUT, gCurrentTime + gChunkTimeoutInterval, path, chunk);
	}
	else {
		PHost source_host = path.from->host();
		double latency = source_host->upload_latency() + download_latency();
		if (source_host->router() != NULL) {
			latency += source_host->router()->network_latency(path);
		}
		double sourceBW = source_host->upload_bandwidth();

		double arrival_time = chunk_arrival_time(h_download_avail, h_pending,
			sourceBW, download_bandwidth()) + latency;
		myassert(arrival_time >= gCurrentTime, "Fault in priority-based host");
		Download download_descriptor(chunk, sourceBW, arrival_time);
		EventManager::schedule_event(CHUNK_DELIVERED, arrival_time, path.from, path.to, chunk, download_descriptor.id);
		PendingIter insertion_point = std::lower_bound(h_pending.begin(), h_pending.end(), download_descriptor,
			download_comparator);
		h_pending.insert(insertion_point, download_descriptor);
		h_download_avail = std::max(0.0, h_download_avail - sourceBW);
	}
}

void PriorityBasedHost::deliver_chunk(PChunk chunk, Path path, size_t id) {

	// Due to time jitter in the event manager or even uncertain handling of identically-timed
	// events (possibly unstable sort), we can't depend on chunk delivery events arriving
	// in exactly the order we planned. So need to allow for a small variance.

	PendingIter it = h_pending.begin();
	while (it != h_pending.end() && it->id != id) {
		++it;
	}
	myassert(it != h_pending.end() && fabs(it->end_time - h_pending.front().end_time) < 2 * gTimeJitter, 
		"Priority sharing algorithm failure #3");
	h_pending.erase(it);

	// Computing amount now available is tricky. This approach might be expensive. 
	// We'll worry about that only if it proves a problem.

	h_download_avail = download_bandwidth();
	for (PendingIter it = h_pending.begin(); it != h_pending.end(); ++it) {
		Bandwidth rate = std::min(it->sourceBW, h_download_avail);
		h_download_avail -= rate;
	}

	Host::deliver_chunk(chunk, path);
}


