#ifndef _PRIORITY_BASED_HOST_H_
/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Defines the PriorityBasedHost subclass of Host. This host type
	allows multiple sharing of a download link in a priority-based
	manner, with bandwidth being allocated on a first-come first-served
	basis.
 */

#define _PRIORITY_BASED_HOST_H_

#include <list>
#include "types.h"
#include "host.h"
#include "globals.h"


/// A PriorityBasedHost allows multiple chunks from lower-speed sources
/// to be downloaded by the host
/// at the same time, but in a prioritised manner. Available download bandwidth
/// is shared among the chunks in a first-come-first-served manner, i.e., the
/// first chunk gets as much bandwidth as it can use (determined by the source's
/// upload bandwidth), the second gets as much of what's left over as it can use,
/// and so on. However, if an attempt is made to send a chunk to a destination
/// host that is saturated, i.e., has zero free bandwidth, the chunk is treated
/// like an error chunk, and is delivered to its destination as a "chunk_timeout"
/// call.

/// A Download is a descriptor of an active chunk download
struct Download {
	size_t id;	// Uniquely identifies the particular download
	static size_t next_id;
	PChunk chunk;
	Bandwidth sourceBW;
	Time start_time;
	Time end_time;

	Download(PChunk chunk, Bandwidth sourceBW, Time end_time) :
		id(next_id++), chunk(chunk), sourceBW(sourceBW),
		start_time(gCurrentTime), end_time(end_time) {}
};

class PriorityBasedHost : public Host {
	std::list<Download> h_pending;  // The list of active downloads to this host.
									// Kept ordered by end_time.
public:
	PriorityBasedHost(Bandwidth uploadBW, Bandwidth downloadBW,
		   Latency upload_latency, Latency download_latency);
	virtual void send_chunk(PChunk chunk, Path path);
	virtual void chunk_scheduled(PChunk chunk, Path path);
	virtual void deliver_chunk(PChunk chunk, Path path, size_t id);
};

#endif // _PRIORITY_BASED_HOST_H_
