/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file stats.cpp
	Implementations of most statistics gathering and printing functionality.
	The definitions of the LinksStats and PeerStats classes are here plus
	other statistics functionality including the statistics methods
	of the Overlay class.
 */
#include <cstdio>
#include <string>
#include "globals.h"
#include "overlay.h"
#include "stats.h"
#include "chunk.h"
#include "misc.h"
#include "peer.h"
#include "link.h"
#include "videostream.h"
#include "myassert.h"

bool PeerStats::s_recording_chunk_arrivals = false;
ChunkID PeerStats::s_first_chunk_to_record = 0;
ChunkID PeerStats::s_last_chunk_to_record = 0;
size_t PeerStats::s_num_delay_counters = 0;
double PeerStats::s_delay_histogram_bar_width = 0.0;

// Constructor and methods for the PeerStats class
// ================================================

PeerStats::PeerStats() {  	// Constructor clears all fields
	clear();
}

void PeerStats::clear() {		// Called for all peers by Overlay.clear()
	s_num_chunks_sent = s_num_chunks_received = s_last_chunk_received
		= s_num_out_of_seq = 0;
	s_chunk_arrivals.clear();
	s_delay_counters.clear();
	s_delay_counters.resize(s_num_delay_counters, 0);
}

void PeerStats::set_first_chunk_to_record(ChunkID chunk) {
	s_first_chunk_to_record = chunk;
}

void PeerStats::set_last_chunk_to_record(ChunkID chunk) {
	s_last_chunk_to_record = chunk;
}

void PeerStats::set_chunk_arrival_recording(bool state) {
		s_recording_chunk_arrivals = state;
}

double PeerStats::delay_histogram_bar_width() {
	return s_delay_histogram_bar_width;
}

size_t PeerStats::num_delay_counters() {
	return s_num_delay_counters;
}

const std::vector<size_t>* PeerStats::delay_counters() {
	return &s_delay_counters;
}

void PeerStats::set_num_delay_counters(size_t n) {
	s_num_delay_counters = n;
	s_delay_histogram_bar_width = gPlayoutDelay / n;
}

// Record the arrival of a chunk at the peer
void PeerStats::record_arrival(PChunk chunk) {
	size_t chunk_id = chunk->id();
	if (chunk_id >= s_first_chunk_to_record && chunk_id <= s_last_chunk_to_record && !chunk->parity()) {
		if (s_num_chunks_received != 0 && s_last_chunk_received > chunk_id) {
			s_num_out_of_seq++;
		}
		s_num_chunks_received++;
		s_last_chunk_received = chunk_id;
		if (s_recording_chunk_arrivals) {
			s_chunk_arrivals[chunk_id] = gCurrentTime;
		}
		size_t bar_num = (size_t) ((gCurrentTime - chunk->initiation_time()) / 
							s_delay_histogram_bar_width);
		bar_num = std::min(bar_num, s_delay_counters.size()-1); // Late chunks go into last bin
		s_delay_counters[bar_num]++;
	}
}

// Record the sending of a chunk by the peer that owns this stats record
void PeerStats::record_send(PChunk pchunk) {
	s_num_chunks_sent++;
}

// Record the fact that a link has been deleted
void PeerStats::record_link_deletion(PPeer to_whom) {
	s_link_deletions.push_back(to_whom);
}

// Return the arrival time of a given chunk, or -1 if it never arrived.
Time PeerStats::arrival_time(ChunkID chunk_id) {
	//return (s_chunk_arrivals.size() <= chunk_id ? -1 : s_chunk_arrivals[chunk_id]);
	
	std::map<ChunkID, Time>::iterator it = s_chunk_arrivals.find(chunk_id);
	return it == s_chunk_arrivals.end() ? -1 : it->second;
}

size_t PeerStats::num_chunks_out_of_seq() {
	return s_num_out_of_seq;
}

size_t PeerStats::num_chunks_sent() {
	return s_num_chunks_sent;
}

size_t PeerStats::num_chunks_received() {
	return s_num_chunks_received;
}

/// The number of chunks lost, between the s_first_chunk_to_record and s_last_chunk_to_record.
size_t PeerStats::num_chunks_lost() {
	return (s_last_chunk_to_record - s_first_chunk_to_record + 1) - s_num_chunks_received;
}
/// Chunk loss ratio (per peer), between the s_first_chunk_to_record and s_last_chunk_to_record.
double PeerStats::chunk_loss_ratio() {
  return 1 - s_num_chunks_received * 1.0 /
    (s_last_chunk_to_record - s_first_chunk_to_record + 1);
}

// Constructor and methods for the LinksStats class
// ================================================

LinksStats::LinksStats() {
	clear();
}

void LinksStats::clear() {
	s_throughput = 0;			// Called for all links by Overlay.clear()
}

size_t LinksStats::throughput() {
	return s_throughput;
}

void LinksStats::inc_throughput() {
	s_throughput++;
}

// Print statistics for all sorts of useful things!
// ================


void Overlay::print_statistics()
{
	size_t first_chunk = o_stats_start;
	size_t last_chunk = o_stats_stop+1;	//not the ID but counting from 1
	size_t num_chunks_considered=last_chunk-first_chunk;

	// Print stats on lost and out-of-sequence chunks

	FILE* fp_chunk_fuori_seq = check_open("OutOfSequenceFile");
	FILE* fp_chunk_persi = check_open("TotalChunkLossesPerPeerFile");
	for(PeerIterator it = first_peer(); it != last_peer(); ++it) {
		fprintf(fp_chunk_persi,"%d\n", (*it)->stats()->num_chunks_lost());
		fprintf(fp_chunk_fuori_seq,"%d\n", (*it)->stats()->num_chunks_out_of_seq());
		//printf("per peer chunk loss ratio: %f\n", (*it)->stats()->chunk_loss_ratio());
	}
	fclose(fp_chunk_fuori_seq);
	fclose(fp_chunk_persi);

	// Print arrival times for all chunks and all peers

	FILE* fp_time_rx = check_open("ChunkArrivalsFile");
	for(size_t i = first_chunk; i < last_chunk && fp_time_rx; i++){
		PChunk chunk = VideoStream::get_chunk(i);
		if (chunk->parity()) {
			//fprintf(fp_time_rx, "# ");
			continue;
		}
		fprintf(fp_time_rx, "%d\t", i);
		Time source_time = chunk->initiation_time();
		for (PeerIterator pit = first_peer(); pit != last_peer(); ++pit) {
			PPeer peer = *pit;
			Time arrival_time = peer->stats()->arrival_time(i);
			if (arrival_time != -1) {
				fprintf(fp_time_rx,"%.5f ", arrival_time);
			}
			else { 
				fprintf(fp_time_rx,"%.5f ",-1.00000);
			}
		}
		fprintf(fp_time_rx, "\n");
	}
	if (fp_time_rx) {
		fclose(fp_time_rx);
	}

	// Print delay times for all chunks and all peers

	FILE* fp_delaychunk1 = check_open("ChunkDelaysFile");

        long gl_chunksarrived = 0;
        long gl_chunkinstances = 0;
        Time gl_totaldelay = 0;
        Time gl_maxdelay = 0;

	double percentile_90 = std::numeric_limits<double>::infinity();
	double percentile_95 = std::numeric_limits<double>::infinity();
	double percentile_99 = std::numeric_limits<double>::infinity();

	for(size_t i = first_chunk; i < last_chunk && fp_delaychunk1; i++){
		PChunk chunk = VideoStream::get_chunk(i);
		if (chunk->parity()) {
			//fprintf(fp_time_rx, "# ");
			continue;
		}
		fprintf(fp_delaychunk1, "%d\t", i);
		Time source_time = chunk->initiation_time();
		for (PeerIterator pit = first_peer(); pit != last_peer(); ++pit) {
			PPeer peer = *pit;
			if (peer->is_source()) continue;	//exclude source
			Time arrival_time = peer->stats()->arrival_time(i);
			gl_chunkinstances++;
			if (arrival_time != -1) {
				Time delay = arrival_time - source_time;
				fprintf(fp_delaychunk1,"%.5f ", delay);
				gl_maxdelay = (delay > gl_maxdelay ) ? delay : gl_maxdelay;
				gl_totaldelay += delay;
				gl_chunksarrived++;
			}
			else { 
				fprintf(fp_delaychunk1,"%.5f ",-1.00000);
			}
		}
		fprintf(fp_delaychunk1,"\n");
	}
	if (fp_delaychunk1) {
		fclose(fp_delaychunk1);
	}

	FILE* fp_perchunkinstance_cdf = check_open("DelayCdfFile");
	if (fp_perchunkinstance_cdf) {

		double *perchunkinstance = new double[gl_chunkinstances];
		for (long l=0;l<gl_chunkinstances;l++){
			perchunkinstance[l]=std::numeric_limits<double>::infinity();;
		}

		long l = 0;
		for(size_t i = first_chunk; i < last_chunk; i++){
			PChunk chunk = VideoStream::get_chunk(i);
			if (chunk->parity()) {
				continue;
			}
			Time source_time = chunk->initiation_time();
			for (PeerIterator pit = first_peer(); pit != last_peer(); ++pit) {
				if ((*pit)->is_source()) continue;
				Time arrival_time = (*pit)->stats()->arrival_time(i);
				if (arrival_time != -1) {
					Time delay = arrival_time - source_time;
					perchunkinstance[l++]=delay;
				}
			}
		}

		fprintf(fp_perchunkinstance_cdf,"#chunkinstance-cdf\tdelay\n");
        	std::sort(perchunkinstance, perchunkinstance + gl_chunksarrived);
		for (l=0;l<gl_chunksarrived;l++){
			if (l==gl_chunkinstances-1 || perchunkinstance[l+1]>perchunkinstance[l]){
				fprintf(fp_perchunkinstance_cdf,"%f\t%f\n",perchunkinstance[l],(double)(l+1)/gl_chunkinstances);
			}
		}
		percentile_90=perchunkinstance[(int)(0.90*gl_chunkinstances)];
		percentile_95=perchunkinstance[(int)(0.95*gl_chunkinstances)];
		percentile_99=perchunkinstance[(int)(0.99*gl_chunkinstances)];

		delete perchunkinstance;
	}
	if (fp_perchunkinstance_cdf) {
		fclose(fp_perchunkinstance_cdf);
	}

        printf("average chunk loss ratio: %f\n", 1 - (double)gl_chunksarrived/gl_chunkinstances);
	printf("average chunk delay for chunks not lost %f\n", gl_totaldelay/gl_chunksarrived);
        printf("maximum chunk delay for chunks not lost %f\n", gl_maxdelay);
        printf("\tTESTresults\t%f\t%f\t%f\t%f\t%f\t%f\n", 1 - (double)gl_chunksarrived/gl_chunkinstances, gl_totaldelay/gl_chunksarrived, gl_maxdelay, percentile_90, percentile_95, percentile_99);

	// Dump of all the peer delay histogram data

	FILE* fp_delay_histograms = check_open("DelayHistogramsFile");
	size_t n = PeerStats::num_delay_counters();
	fprintf(fp_delay_histograms, "%d %d %.3f\n", n, num_peers(),
		PeerStats::delay_histogram_bar_width());
	for (PeerIterator it = first_peer(); it != last_peer(); ++it) {
		fprintf(fp_delay_histograms, "%d %d %d\n", (*it)->id(), 
			(*it)->get_class(), (*it)->stats()->num_chunks_lost());
		const std::vector<size_t>* cntrs = (*it)->stats()->delay_counters();
		for (size_t i = 0; i < n; i++) {
			fprintf(fp_delay_histograms, "%d ", (*cntrs)[i]);
		}
		fprintf(fp_delay_histograms,"\n");
	}
	fclose(fp_delay_histograms);

	/* Creating the file with upload bandwidth and class and churn status of each peer*/

	FILE* fp_upload_cap = check_open("UploadRatePeerFile");
	for (PeerIterator it = first_peer(); it != last_peer(); ++it) {
		fprintf(fp_upload_cap,"%d %.5f %d %d\n", (*it)->id(), (*it)->upload_bandwidth(),
			(*it)->get_class(), (*it)->get_churner());
	}
	fclose(fp_upload_cap);

	/* Creating the file with s_throughput of each peer*/

	FILE* fp_upload_total = check_open("UploadTotalPeerFile");
	FILE* fp_throughput = check_open("ThroughputFile");
	for (PeerIterator pit = first_peer(); pit != last_peer(); ++pit) {
		PPeer source = *pit;
		size_t totalupload = 0;
		for (LinkIterator it = source->first_out_link(); it != source->last_out_link(); ++it) {
			PLink pLink = *it;
			PPeer dest = pLink->to();  // Target of this link
			size_t s_throughput = pLink->stats()->throughput();
			totalupload += s_throughput;
			fprintf(fp_throughput,"%d %d %d\n", source->id(), dest->id(), s_throughput);		 
		}
		fprintf(fp_upload_total,"%d %d\n", source->id(), totalupload);
	}		
	fclose(fp_upload_total);
	fclose(fp_throughput);

	return;
}
