#!/bin/bash
#********************************************************************
# Copyright (C) 2008  University of Trento
#
# Author:    Csaba Kiraly
#  	     surname (at) disi.unitn.it
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#********************************************************************

# Usage: set paramters in config.script.txt, config.sh or in a separate testcase file script,
# then run it with the command line option below.
# A results file (${prefix}results.txt) will be created, with one line for each simulation run.
# The results line first contains the input paramters then the string "TESTresults", then output values.

function usage () {
   echo "Usage: $PROGNAME -d results_dir -t testcase     -p filename_prefix [-s src_dir] [-c]"
   echo "Usage: $PROGNAME -d results_dir -T testcasefile -p filename_prefix [-s src_dir] [-c]"
   echo "        -t test: name of the testcase to execute from test.sh or the file with the testcase"
   echo "        -T testfile: name of the file describing a testcase (parameters in the file override -t)"
   echo "        -d dir: directory for result files"
   echo "        -p prefix: filename prefix of result files"
   echo "        -s dir: directory with sources, used also for recompile"
   echo "        -c: copy sources to resultsdirectory/src and compile there; implies -b"
   echo "        -b: run in background"
   echo "        -P: calcualte percentiles as well"
   echo "        -r: resume simulations, reusing previous results"
   echo "        -F: regenerate gnuplot files and figures only"
   echo "        -h: this help"
   exit $1
}
# examples:
#
# ./test.sh -d results -t numnodes -p prefix
# Starts recompiling the program in place, and puts results in the "./results" directory, with the given prefix.
# It runs the "numnodes" test case, i.e. it studies performance as a funcion of the number of nodes.
# A copy of all parameters will be saved in the "results" folder.
# A script "prefix_rerun.sh" will also be created in the "results" directory, keeping track of the command line options.
#
# ./test.sh -d results/001 -p 001_ -t numnodes -p prefix -c
# -c means that it will make a snapshot of the sources under "results/001/src"
# These sources will be used for compilation which will also be done in the new folder, 
# thus it will not interfere with other simulations or modifications to the source.
# The sript will go into background, immune to hang-ups (like nohup).
#
# ./test.sh -d results/002 -T TESTCASE-2 -p prefix
# Starts recompiling the program in place, and puts results in the "./results/002" directory.
# It runs the testcase as defined in the file TESTCASE-2
#
# ./test.sh -d results -t chunkdelay-cdf -p prefix
# Starts recompiling the program in place, and puts results in the "./results" directory.
# It runs the "chunkdelay-cdf" test case, i.e. it genates a CDF for all given combination of chunk and peer scheduling algorithms
#

scriptspath=`dirname $0`

#==================================================================
# Define line format for results file. All the variables below will be saved for each simulation run
LINEHEAD=\
'$RUN_UID\t$CYCLE\t$SEED'\
'\t$NUM_PEERS\t$DEGREE'\
'\t$VIDEORATE'\
'\t$MEANBW\t$PEERSCLASS1\t$MEANCLASS1\t$PEERSCLASS2\t$MEANCLASS2\t$PEERSCLASS3\t$MEANCLASS3\t$PEERSCLASS4\t$MEANCLASS4\t$DOWNLOAD_BW\t$VARIANCERATE'\
'\t$PROB_CHUNK_ERROR\t$CHUNK_TIMEOUT'\
'\t$NUM_CHUNKS\t$FECK\t$FECN'\
'\t$SCHED_ORDER\t$CSCHED\t$PSCHED\t$WIN_SIZE\t$PLAYOUT_DEADLINE'\
'\t$VAR_I\t$VAR_J\t$VAR_K'\
''
# This is just a descriptive header line for the results file
HEADINGLINE=\
'#runuid\tcycle\tseed'\
'\tpeers\tdegree'\
'\tvideorate'\
'\tmeanbw\tpeercl1\tbw1\tpeercl2\tbw2\tpeercl3\tbw3\tpeercl4\tbw4\tdown\tvarrate'\
'\tchunkerr\tchtimeout'\
'\tchunks\tfeck\tfecn\tschord\tcsched\tpsched\twinsize\tplayout'\
'\ti\tj\tk'\
'\tTESTresults\tlossratio\tavgdelay\tmaxdelay\t90percentile\t95percentile\t99percentile'\
''

#==================================================================
# Include helper scripts
source $scriptspath/utils.sh
source $scriptspath/config.sh
source $scriptspath/plot.sh
source $scriptspath/backend-p2ptvOO.sh

#==================================================================
# Process command-line parameters
PROGNAME=$0
CMDLINE="$0 $*"

while getopts "d:p:cs:t:T:brhPF" opt; do
   case $opt in

   d )  PREFIXDIR=$OPTARG ;;
   p )  PREFIX=$OPTARG ;;
   s )  SRCDIR=$OPTARG ;;
   c )  COPYSOURCE=1;;
   t )  TESTCASE=$OPTARG ;;
   T )  TESTCASEFILE=$OPTARG ;;
   b )  BACKGROUND=1;;
   P )  PERCENTILE=1;;
   r )  RESUME=1;;
   F )  FIGURESONLY=1;;
   h )  usage 0 ;;
   \?)  usage 1 ;;
   esac
done

# Verify obligatory and extra parameters
shift $(($OPTIND - 1))
if [[ $1 || ! ( $TESTCASE || $TESTCASEFILE ) || ! $PREFIXDIR || ! $PREFIX ]]; then
  usage 1
fi

[[ $SRCDIR ]] || SRCDIR=`pwd`

#convert to abolute path
SRCDIRABS=`cd $SRCDIR; pwd`

TESTOUT=${PREFIX}results.txt

#create directory for results
mkdir -p $PREFIXDIR
PREFIXDIRABS=`cd $PREFIXDIR; pwd`

#copy sources if required (-c)
if [[ $COPYSOURCE ]]; then
  copy_source "$PREFIXDIR" "$TESTCASEFILE"
  run_in_dir "$PREFIXDIR" "$PREFIX" "$TESTCASE" "$PERCENTILE" "$TESTCASEFILE"
else 
  if [[ $TESTCASEFILE ]]; then
    cp $TESTCASEFILE $PREFIXDIR
  fi
  cd $PREFIXDIR || exit 1
fi

# Save code with parameters for later reference.
# Enable this only if you rename test.sh for experimentation. 
# In this case git won't copy the file, so this explicit copy is needed
#cp $SRCDIR/$0 ${PREFIX}`basename $0`

# Redirect stdout(1) and stderr(2) to logfiles and stdout.
# By default, the logfiles are compressed on the fly using gzip. The previous
# non-compressed version is still available (uncomment the appropriate lines to use it).
if [[ $BACKGROUND ]]; then
  #exec 2>${PREFIX}.err 1>${PREFIX}.out                                              # plain unzipped output
  exec 2> >(gzip -c >${PREFIX}.err.gz) 1> >(gzip -c >${PREFIX}.out.gz)               # gzipped output (default)
else
  #exec 2> >(tee ${PREFIX}.err) 1> >(tee ${PREFIX}.out)                              # plain unzipped output
  exec 2> >(tee >(gzip -c >${PREFIX}.err.gz)) 1> >(tee >(gzip -c >${PREFIX}.out.gz)) # gzipped output (default)
fi

#set parameters for the given test case (-t)
if [[ $TESTCASE ]]; then
  configure_testcase $TESTCASE
fi
# TESTCASEFILE overrides settings from the TESTCASE (-T)
if [[ $TESTCASEFILE ]]; then
  . $TESTCASEFILE || exit 1
fi

#create gnuplot files
if [[ ! $CURVEVARS ]]; then
  CURVEVARS="CSCHED PSCHED SCHED_ORDER" #variables to use to differentiate curves
fi
plotfiles "${CURVEVARS}"

#handle -F parameter
if [[ $FIGURESONLY ]] ; then
  plot
  exit 1
fi

#create results file
if [[ ! $RESUME || ! -f $TESTOUT ]]; then
  echo -e $HEADINGLINE >$TESTOUT
fi

#save command line parameters
# TODO convert to relative path to improve usability
if [[ ! -f ${PREFIX}rerun.sh ]]; then
  echo $CMDLINE -r >${PREFIX}rerun.sh
  chmod u+x ${PREFIX}rerun.sh
fi
if [[ ! -f ${PREFIX}replot.sh ]]; then
  echo $CMDLINE -F >${PREFIX}replot.sh
  chmod u+x ${PREFIX}replot.sh
fi
#write PID and create kill script
myPID=$$
echo $myPID >${PREFIX}.PID
echo "[[ `hostname` == \`hostname\` ]] || { echo 'you should better kill on `hostname`'; exit 1; } " >${PREFIX}kill.sh
echo "kill $myPID" >>${PREFIX}kill.sh
echo "rm ${PREFIX}.PID ${PREFIX}kill.sh" >>${PREFIX}kill.sh

chmod u+x ${PREFIX}kill.sh

#start the big cycle
RUN_UID_COUNTER=1

echo NUM_PEERS: $NUM_PEERSs

#initialize backend
backend_init $SRCDIRABS

function runonecycle {
  date
  configure_init $SRCDIRABS
  configure_network $SRCDIRABS
  configure_scheduler $SRCDIRABS
  configure_other $SRCDIRABS
  #RUN_UID=${RUN_UID_COUNTER}
  RUN_UID=tmp
  mkdir -p $RUN_UID

  eval LINEHEAD2=\"$LINEHEAD\"
  if [[ $RESUME && `awk "/$LINEHEAD2/" $TESTOUT` ]] ; then
    echo "$LINEHEAD2 skipped by resume"
    return 1
  fi

  echo -e  $LINEHEAD2
  RESULTSLINE=$LINEHEAD2

  run $SRCDIRABS $SEED $RUN_UID
  let RUN_UID_COUNTER+=1
}

#start the big cycle. Cycle through the given parameters, calling functions in between
cyclevariables \
  CYCLE SEED \
  postcall plot \
  VIDEORATE \
  precall video_init \
  MEANBW PEERSCLASS1 MEANCLASS2 MEANCLASS1 PEERSCLASS2 MEANCLASS3 PEERSCLASS3 MEANCLASS4 PEERSCLASS4 DOWNLOAD_BW VARIANCERATE NUM_PEERS DEGREE \
  PROB_CHUNK_ERROR CHUNK_TIMEOUT \
  NUM_CHUNKS FECK FECN \
  WIN_SIZE PLAYOUT_DEADLINE SCHED_ORDER CSCHED PSCHED \
  VAR_I VAR_J VAR_K \
  precall runonecycle

#remove PID file and kill script
rm ${PREFIX}.PID ${PREFIX}kill.sh
