/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file 
    Implementation of the TweakedOldSimHost subclass of the Host class.
	This is a variant of the OldSimHost class that allows a highspeed
	uploader to complete before its lower-speed downloader has received
	the chunk.
 */

#include "myassert.h"
#include "tweakedoldsimhost.h"
#include "globals.h"
#include "path.h"
#include "event.h"
#include "router.h"
#include "peer.h"
#include "misc.h"

/// A TweakedOldSimHost is like the OldSimHost except that when a high-bandwidth
/// host is feeding a low-bandwidth one, the source completes its send in a time
/// determined by its own upload bandwidth, and gets called back at that time, allowing
/// it to schedule other chunks.
TweakedOldSimHost::TweakedOldSimHost(Bandwidth uploadBW, Bandwidth downloadBW,
		   Latency upload_latency=0, Latency download_latency=0)
		   : OldSimHost(uploadBW, downloadBW, upload_latency, download_latency) {}

void TweakedOldSimHost::send_chunk(PChunk chunk, Path path) {
	double source_send_time = chunk->size() / upload_bandwidth();
	double source_delay = source_send_time + upload_latency();
	double source_delivery_time = gCurrentTime + source_delay;
	EventManager::schedule_event(CHUNK_SENT, source_delivery_time, path.from, path.to, chunk);
	if (router() != NULL) {
		router()->sending_chunk(chunk, path);  /// Notify router of incoming chunk
	}
	path.to->host()->chunk_scheduled(chunk, path);
}

void TweakedOldSimHost::chunk_scheduled(PChunk chunk, Path path) {
	double effective_bandwidth = std::min(path.from->upload_bandwidth(), download_bandwidth());
	double transmission_delay = chunk->size() / effective_bandwidth;
	double latency = path.from->host()->upload_latency() + download_latency();
	if (router() != NULL) {
		latency += router()->network_latency(path);
	}
	if (!is_infinite(download_bandwidth())) {
		h_download_avail = 0;
	}
	double arrival_time = gCurrentTime + transmission_delay + latency;
	EventManager::schedule_event(CHUNK_DELIVERED, arrival_time, path, chunk);
}

