/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Implementation of the video stream class, which manages an internal
	vector of chunks called the "chunk list" and delivers them to the
	source peer of the overlay at regular intervals. The class also includes
	a callback mechanism that allows clients to schedule one-off backbacks\
	just prior to distribution of a specified chunk or repeating callbacks
	at regular chunk intervals.
 */

#include <vector>
#include <list>
#include <algorithm>
#include "types.h"
#include "misc.h"
#include "videostream.h"
#include "chunk.h"
#include "overlay.h"
#include "globals.h"
#include "peer.h"
#include "event.h"

std::vector<PChunk> VideoStream::chunk_list;
size_t VideoStream::current_chunk_id = 0;
size_t VideoStream::num_chunks = 0;
size_t VideoStream::num_chunks_w_par = 0;
std::list<VideoStreamCallback> VideoStream::callbacks;
size_t VideoStream::fec_n = 0;	// blocksize with parity chunks
size_t VideoStream::fec_k = 0;	// blocksize

/// Get the number of chunks in the video stream
size_t VideoStream::get_num_chunks( bool w_par) {
	return w_par ? num_chunks_w_par : num_chunks;
}

/// Get the number of the next chunk to send
size_t VideoStream::get_next_chunk_num() {
	return current_chunk_id;
}

/// Set the number of chunks in the video stream
void VideoStream::set_num_chunks(size_t num_chunks) {
	VideoStream::num_chunks = VideoStream::num_chunks_w_par = num_chunks;
}

void VideoStream::add_callback(void (*func)(size_t chunk_num), size_t chunk_num, bool repeats) {
	callbacks.push_back(VideoStreamCallback(func, chunk_num, repeats));
}

// Predicate for use with delete_callback method
bool matches(VideoStreamCallback cb, void (*func)(size_t)) {
	return cb.callback_func == func;
}

void VideoStream::delete_callback(void (*func)(size_t chunk_num)) {
	callbacks.remove_if(std::bind2nd(std::ptr_fun(matches), func));
}

/// Give the next chunk to the source node.
void VideoStream::next_chunk(POverlay po) {

	// Firstly, call any callbacks matching the number of the chunk we're about to send,
	// except that callbacks aren't called for chunk 0.
	// Non-repeating callbacks are deleted after one use.
	if (current_chunk_id != 0) {
		for (std::list<VideoStreamCallback>::iterator it = callbacks.begin(); it != callbacks.end();) {
			if (current_chunk_id % it->chunk_number == 0)  {
				it->callback_func(current_chunk_id);
				if (!it->repeats) {
					it = callbacks.erase(it);
				}
				else {
					++it;
				}
			}
			else {
				++it;
			}
		}
	}

	std::vector<PChunk> to_send;  //vector of chunks to send to the source

	//FEC here
	if (fec_n && (current_chunk_id % fec_n) == fec_k-1) { //we put the FEC chunks before the last chunk fo the block, to help LUc like strategies take the good one
		for (int i=fec_k-1; i<fec_n-1; i++) {	// generate parity chunks
			PChunk chunk = new ChunkInstance(current_chunk_id++, gCurrentTime, gChunkSize, 'I', true);	//TODO: set chunk size as needed
			chunk->set_dl(chunk->id());
			num_chunks_w_par++;	// parity chunks are extra!
		        if (chunk_trace) printf("T = %lf, chunk %d (FEC) added to source's buffer, %ld\n", gCurrentTime, chunk->id(), gSeed);
	                chunk_list.push_back(chunk);	
                        to_send.push_back(chunk);
		}
	}

	PChunk chunk = new ChunkInstance(current_chunk_id++, gCurrentTime, gChunkSize, 'I');
	chunk->set_dl(chunk->id());
	chunk_list.push_back(chunk);	
        to_send.push_back(chunk);

	if (chunk_trace) printf("T = %lf, chunk %d added to source's buffer, %ld\n", gCurrentTime, chunk->id(), gSeed);

	double deltaT = gChunkSize/(float)gVideoRate;  // Interchunk time at source
	if (current_chunk_id < num_chunks_w_par) {
		EventManager::schedule_event(CHUNK_INITIATION, gCurrentTime + deltaT, NULL, NULL, NULL);
	} else {
		EventManager::schedule_event(SIMULATION_END, gCurrentTime + deltaT + gPlayoutDelay, NULL, NULL, NULL);
	}

	for (std::vector<PChunk>::iterator it = to_send.begin(); it != to_send.end(); it++) {
		po->source_peer()->receive_chunk(*it, NULL);
	}
}


PChunk VideoStream::get_chunk(size_t chunk_id)
{
	return chunk_list[chunk_id];
}

void VideoStream::clear() {
	while (chunk_list.size() > 0) {
		PChunk last = chunk_list.back();
		delete last;
		chunk_list.pop_back();
	}
}
