/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: MMarco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file window.cpp
	Implementations of the window-handling functions declared in window.h
 */

#include <cmath>
#include "myassert.h"
#include "assert.h"
#include "window.h"
#include "random.h"
#include "misc.h"
#include "globals.h"
#include "chunk.h"
#include "chunkinstance.h"
#include "videostream.h"
#include "stats.h"

/// Insert the given chunk into the window. Ignore chunks older than the playout delay.
/// Delete any chunks older than the playout delay. Delete the corresponding chunk
/// from the pending set if it's there.
void Window::insert(PChunk pchunk, PPeerStats stats)
{
	Time dead_time = gCurrentTime - gPlayoutDelay;
	if (chunks.find(pchunk->id()) != chunks.end()) {	// Check for duplicates
	  printf("Duplicate received\n");	
	} else if (pchunk->initiation_time() > dead_time) { 
		chunks[pchunk->id()] = pchunk;
		fec_reconstruct(pchunk,stats);
	}

	while (!chunks.empty() && chunks.begin()->second->initiation_time() < dead_time) {
		useless.erase(chunks.begin()->second->id());  // Only keep status of chunks in window
		chunks.erase(chunks.begin()->second->id());
	}
	pending.erase(pchunk->id());
}



/// Reconstruct chunks in the window using FEC. Only checks the block of the given chunk.
void Window::fec_reconstruct(PChunk pchunk, PPeerStats stats)
{
	int fec_n = VideoStream::get_fec_n();
	int fec_k = VideoStream::get_fec_k();

	//reconstruct chunk block using FEC if available.
	if (fec_n) {
		int blockID = pchunk->id() / fec_n;	// the block effected by the reception
	        int received=0;	// number of chunks already received from the block
		for (int i=blockID*fec_n; i<(blockID+1)*fec_n; i++) {
			if (chunks.find(i) != chunks.end()) {received++;}
	        }
	        if (received>=fec_k) {	//reconstruct
			for (int i=blockID*fec_n; i<(blockID+1)*fec_n; i++) {
				if (chunks.find(i) == chunks.end()) {
					PChunk pchunki = VideoStream::get_chunk(i);
					chunks[i] = pchunki; 
					stats->record_arrival(pchunki);
				}
	        	}
	        }
	}
}

/// Clear the window
void Window::clear()
{
	chunks.clear();
	pending.clear();
	useless.clear();
}

/// Latest chunk
const PChunk Window::latest() {
	PChunk l = NULL;
	if (!chunks.empty()) l=*begin();
	return l;
}

/// Latest chunk, including pending ones
const PChunk Window::latest_w_pending() {
	PChunk l = latest();

	for (CHUNK_MAP::iterator it = pending.begin(); it != pending.end(); ++it) {
            if (!l || l->id() < (*it).second->id()) {
                l = (*it).second;
	    }
	}
	return l;
}



